/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.binding.soap.listener.incoming;

import java.io.IOException;
import java.io.OutputStream;
import java.net.URI;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Logger;

import javax.jbi.component.ComponentContext;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.apache.axis2.AxisFault;
import org.apache.axis2.description.AxisService;
import org.apache.axis2.engine.AxisConfiguration;
import org.objectweb.petals.component.framework.util.Extensions;
import org.objectweb.petals.component.framework.util.WSDLHelper;
import org.objectweb.petals.jbi.descriptor.Consumes;
import org.w3c.dom.Document;

/**
 * The {@link PetalsAxisService} extends an {@link AxisService}. It is
 * registered during SU deployment into the {@link AxisConfiguration}.
 *
 * This class override the printWSDL method to dynamically generate the WSDL
 * from the JBI container.
 *
 * @author Christophe HAMERLING - eBMWebSourcing
 *
 */
public class PetalsAxisService extends AxisService {

    /**
     * The JBI component context
     */
    private ComponentContext componentContext;

    /**
     * The consumes JBI extensions used in this AxisService
     */
    private Consumes consumes;

    /**
     * The consumes cdk extensions
     */
    private Extensions consumesCDKExtensions;

    /**
     * The local URL where this service is exposed (http://HOST:PORT/PREFIX)
     */
    private String localAddress;

    /**
     * Map of operation MEPs
     */
    private Map<QName, URI> operationMep;

    /**
     * The logger
     */
    private Logger logger;

    /**
     * The client service name prefix
     */
    public static final String OUTGOING_SERVICE_CLIENT_PREFIX = "OutgoingWSClient";

    /**
     * Creates a new instance of {@link PetalsAxisService}
     *
     * @param name
     * @param componentContext
     * @param jbiEndpoint
     * @param extensions
     */
    public PetalsAxisService(final String name,
            final ComponentContext componentContext, final Consumes consumes,
            final Extensions consumesCDKExtensions, final String localAddress,
            final Logger logger) {
        super(name);
        this.componentContext = componentContext;
        this.consumes = consumes;
        this.localAddress = localAddress;
        this.logger = logger;
        this.consumesCDKExtensions = consumesCDKExtensions;
        this.operationMep = new HashMap<QName, URI>();
    }

    /*
     * (non-Javadoc)
     *
     * @see org.apache.axis2.description.AxisService#printWSDL(java.io.OutputStream,
     *      java.lang.String, java.lang.String)
     */
    @Override
    public void printWSDL(final OutputStream out, final String requestIP)
            throws AxisFault {
        this.logger.fine("Print WSDL");

        try {
            Document doc = getServiceDefinition();
            if (doc == null) {
                this.printWSDLError(out,
                                "WSDL description can not been retrieved from JBI endpoint");
            } else {
                WSDLHelper.printWSDL(doc, out);
            }

        } catch (Exception e) {
            throw new AxisFault("Server", e);
        }
    }

    /*
     * (non-Javadoc)
     *
     * @see org.apache.axis2.description.AxisService#printWSDL2(java.io.OutputStream,
     *      java.lang.String)
     */
    @Override
    public void printWSDL2(final OutputStream out, final String requestIP)
            throws AxisFault {
        this.printWSDL(out, requestIP);
    }

    /**
     * Print the WSDL error on the output stream
     *
     * @param out
     * @throws AxisFault
     */
    private void printWSDLError(final OutputStream out, String message) {
        String wsdlntfound = "<error><description>Unable to get WSDL for this service</description><reason>"
                + message + "</reason></error>";
        this.printMessage(out, wsdlntfound);
    }

    /**
     * Print a message to the output stream
     *
     * @param out
     * @param str
     * @throws AxisFault
     */
    private void printMessage(final OutputStream out, final String str) {
        try {
            out.write(str.getBytes());
            out.flush();
        } catch (IOException e) {
        } finally {
            try {
                out.close();
            } catch (IOException e) {
                // Do nothing
            }
        }
    }

    /**
     * Get the service definition document (WSDL)
     *
     * @return
     * @throws Exception
     */
    public Document getServiceDefinition() throws Exception {
        this.logger.fine("Get the service definition");

        Document result = null;
        ServiceEndpoint ep = getServiceEndpoint();
        if (ep != null) {
            result = this.componentContext
                    .getEndpointDescriptor(getServiceEndpoint());

            if (result != null) {
                result = WSDLHelper.replaceAddressServiceIgnoreNS(result,
                        this.localAddress, getName());

            }
        }
        return result;
    }

    /**
     * Get the service endpoint. Try to retrieve if the endpoint name and
     * service name has been specified, else try to get it if only the interface
     * has been specified.
     *
     * @return
     */
    public ServiceEndpoint getServiceEndpoint() {
        ServiceEndpoint jbiEndpoint = null;

        if (consumes.getEndpointName() != null
                && consumes.getServiceName() != null) {
            // get from EP and service values
            jbiEndpoint = this.componentContext.getEndpoint(this.consumes
                    .getServiceName(), this.consumes.getEndpointName());
        } else if (this.consumes.getInterfaceName() != null) {
            // get from ITF value
            ServiceEndpoint[] endpoints = this.componentContext
                    .getEndpoints(this.consumes.getInterfaceName());
            if (endpoints != null && endpoints.length > 0) {
                jbiEndpoint = endpoints[0];
            }
        }
        this.logger.fine("JBI Service endpoint " + jbiEndpoint);
        return jbiEndpoint;
    }

    /**
     * @return the actionMep
     */
    public Map<QName, URI> getOperationMep() {
        return this.operationMep;
    }

    /**
    *
     * @return the consumes related cdk extensions
    */
    public Extensions getConsumesCDKExtensions() {
        return this.consumesCDKExtensions;
   }

   /**
    *
    * @return
    */
   public Consumes getConsumes() {
       return this.consumes;
   }
}
