/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: Axis2BCListener.java 154 19 avr. 2006 wjoseph $
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.binding.soap.listener.incoming;

import java.net.URI;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.InOptionalOut;
import javax.jbi.messaging.InOut;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessageExchangeFactory;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;
import javax.xml.transform.Source;

import org.apache.axiom.om.OMElement;
import org.apache.axiom.soap.SOAPBody;
import org.apache.axiom.soap.SOAPEnvelope;
import org.apache.axiom.soap.SOAPFactory;
import org.apache.axis2.AxisFault;
import org.apache.axis2.Constants;
import org.apache.axis2.context.MessageContext;
import org.apache.axis2.description.Parameter;
import org.apache.axis2.engine.AxisEngine;
import org.apache.axis2.receivers.AbstractMessageReceiver;
import org.apache.axis2.util.MessageContextBuilder;
import org.objectweb.petals.binding.soap.listener.SUProperties;
import org.objectweb.petals.binding.soap.util.Marshaller;
import org.objectweb.petals.component.framework.AbstractComponent;
import org.objectweb.petals.component.framework.ExtensionsConstants;
import org.objectweb.petals.component.framework.MEPConstants;
import org.objectweb.petals.component.framework.mbean.PerformanceNotification;
import org.objectweb.petals.component.framework.util.ExchangeImpl;
import org.objectweb.petals.component.framework.util.Extensions;
import org.objectweb.petals.component.framework.util.StringHelper;
import org.objectweb.petals.jbi.descriptor.Consumes;
import org.w3c.dom.Document;

/**
 * The message receiver used by the binding component. The SOAP message is
 * forwarded to the JBI endpoint.
 *
 * @version $Id$
 * @since Petals 1.0
 * @author alouis,wjoseph,chamerling - EBM Websourcing
 *
 */
public class PetalsReceiver extends AbstractMessageReceiver {

    protected static final String ERROR_WRONG_MESSAGE_STATUS = "JBI message has wrong status";

    public static final String FAULT_SERVER = "Server";

    public static final String FAULT_CLIENT = "Client";

    protected DeliveryChannel channel;

    protected AbstractComponent component;

    protected MessageExchangeFactory exchangeFactory;

    protected Marshaller marshaller = Marshaller.getInstance();

    protected Logger logger;

    /**
     * Creates a new instance of PetalsReceiver
     *
     * @param context
     * @param channel
     * @param bindingSUM
     * @param log
     */
    public PetalsReceiver(AbstractComponent component, DeliveryChannel channel,
            Logger log) {
        super();
        this.channel = channel;
        this.component = component;
        this.logger = log;
        this.exchangeFactory = channel.createExchangeFactory();
    }

    /*
     * (non-Javadoc)
     *
     * @see org.apache.axis2.engine.MessageReceiver#invokeBusinessLogic(org.apache.axis2.context.MessageContext)
     */
    public final void invokeBusinessLogic(final MessageContext msgContext) throws AxisFault {
        logger.log(Level.FINE, "Receiving a SOAP message from Axis2 engine");

        // Performance notification creation
        PerformanceNotification performanceNotification = this.component
                .createPerformanceNotificationReceivingFromOutside(msgContext
                        .getTo().getAddress(), null);

        try {
            // FIX NPE (Axis2 bug)
            Parameter param = new Parameter(
                    Constants.Configuration.SEND_STACKTRACE_DETAILS_WITH_FAULTS,
                    Boolean.TRUE);
            if (msgContext.getAxisOperation() != null) {
                msgContext.getAxisOperation().addParameter(param);
            } else if (msgContext.getAxisService() != null) {
                msgContext.getAxisService().addParameter(param);
            } else if (msgContext.getAxisServiceGroup() != null) {
                msgContext.getAxisServiceGroup().addParameter(param);
            }

            MessageContext outMsgContext = MessageContextBuilder
                    .createOutMessageContext(msgContext);
            outMsgContext.getOperationContext()
                    .addMessageContext(outMsgContext);

            // send message to JBI endpoint
            this.process(msgContext, outMsgContext, performanceNotification);

            // send the response if required
            if (outMsgContext.getEnvelope() != null) {
                AxisEngine.send(outMsgContext);
            }
        } finally {
            // Performance notification sending
            this.component.sendPerformanceNotification(performanceNotification);
        }
    }

    /**
     * Sends the SOAPEnvelope of the request message directly to a JBI service
     * and operation on the JBI container. The response message from the JBI
     * service invocation is transmitted as an outgoing message.
     *
     * @param inMessage
     *            MessageContext for incoming message.
     * @param outMessage
     *            MessageContect fot outgoing message.
     * @param performanceNotification
     *            The performance notification whose the UID will be forwarded
     *            to the service provider.
     * @throws AxisFault
     *             error while processing request, send the JBI Exchange,
     *             processing the JBI response, or if the JBI response is a
     *             Fault.
     */
    protected void process(final MessageContext inContext,
            final MessageContext outContext,
            final PerformanceNotification performanceNotification)
            throws AxisFault {
        logger.log(Level.FINE, "Processing the incoming SOAP message");

        // retrieve Axis service from context
        PetalsAxisService axisService = null;
        if (inContext.getAxisService() instanceof PetalsAxisService) {
            axisService = (PetalsAxisService) inContext.getAxisService();
        } else {
            throw new AxisFault(
                    "Can not retrieve a valid PEtALS Axis service from context", FAULT_CLIENT);
        }

        // Retrieve message operation and service
        QName jbiOperation = this.retrieveSOAPOperation(inContext);
        if (jbiOperation == null) {
            throw new AxisFault("Operation not found in message context", FAULT_CLIENT);
        }

        // get timeout
        long timeout = SUProperties.retrieveTimeOut(axisService.getConsumesCDKExtensions()
                .getPetalsExtensions(), 0);

        // create the message exchange
        MessageExchange exchange = null;
        try {
            exchange = createMessageExchange(inContext, jbiOperation,
                    axisService, performanceNotification);
        } catch (AxisFault e) {
            String errorMsg = "Error while transforming SOAP request to JBI MessageExchange";
            logger.log(Level.SEVERE, errorMsg, e);
            throw new AxisFault(errorMsg, FAULT_SERVER, e);
        }

        // send the message exchange through JBI
        // As the SOAP/HTTP request may wait for a response,
        // use the sendSynchrone MEP
        boolean sent = false;
        PerformanceNotification l_performanceNotification = null;
        try {
            // Performance notification creation
            l_performanceNotification = this.component
                    .createPerformanceNotificationSendingToNMR(inContext
                            .getTo().getAddress(), new ExchangeImpl(exchange));

            sent = channel.sendSync(exchange, timeout);

        } catch (MessagingException e) {
            String errorMsg = "Error while sending message through JBI NMR.";
            logger.log(Level.SEVERE, errorMsg, e);
            throw new AxisFault(errorMsg, FAULT_SERVER, e);
        } finally {
            // Performance notification sending
            this.component
                    .sendPerformanceNotification(l_performanceNotification);
        }

        if (!sent) {
            throw new AxisFault(
                    "A timeout occurs calling the consumed service.", FAULT_SERVER);
        }

        // process the response received from JBI NMR. The out context is
        // updated
        this.processJBIResponse(exchange, outContext, getSOAPFactory(inContext));
    }

    /**
     * Create the message exchange
     *
     * @param inContext
     * @param operationName
     * @param axisService
     * @param performanceNotification
     *            The performance notification whose the UID will be forwarded
     *            to the service provider.
     * @return
     * @throws AxisFault
     */
    protected MessageExchange createMessageExchange(MessageContext inContext,
            final QName operationName, final PetalsAxisService axisService,
            PerformanceNotification performanceNotification) throws AxisFault {
        ExchangeImpl exchangeWrapper = null;

        // retrieve MEP for operation
        // MEPConstants mep = getPattern(operationName, axisService);

        try {

            final Source source = marshaller.createSourceFromSOAPEnvelope(
                    inContext.getEnvelope(), SUProperties
                            .retrieveRemoveRootElement(axisService.getConsumesCDKExtensions()
                                    .getPetalsExtensions(), false));

            // create the message exchange from the extensions. If the MEP is
            // not defined, default will be IN_OPTIONAL_OUT one.
            MessageExchange exchange = createMessageExchange(axisService.getConsumesCDKExtensions());
            exchangeWrapper = new ExchangeImpl(exchange);

            exchangeWrapper.setInMessageContent(source);
            exchangeWrapper.setOperation(operationName);

            // Set the service name, endpoint name and interface name. The
            // Router will find the JBI endpoint to contact with these
            // informations
            exchangeWrapper.setService(axisService.getConsumes()
                    .getServiceName());
            exchangeWrapper.setInterfaceName(axisService.getConsumes()
                    .getInterfaceName());
            exchangeWrapper.setEndpoint(this.findJBIEndpoint(axisService
                    .getConsumes()));

            // add SOAP attachments to normalized message
            marshaller.copySOAPAttachmentsToNM(inContext.getAttachmentMap(),
                    exchangeWrapper.getInMessage());

            // Get the options from the SOAP message and put them into the JBI
            // message
            marshaller.copySOAPPropertiesToME(inContext, exchangeWrapper);

            // Set the performance notification UID, if needed
            if (performanceNotification != null) {
                performanceNotification
                        .setUIDInMessageExchange(exchange);
            }

        } catch (Exception e) {
            throw AxisFault.makeFault(e);
        }

        return exchangeWrapper.getMessageExchange();
    }

    /**
     * Return the JBI ServiceEndpoint according to : <br>{
     * axisService.getTargetNamespace() } axisService.getName() <br>
     * axisService.getEndpoint()
     *
     * @param consumes
     * @return the corresponding endpoint, null if no JBI endpoint has been
     *         found
     */
    protected ServiceEndpoint findJBIEndpoint(final Consumes consumes) {
        ServiceEndpoint endpoint = null;
        if (consumes.getEndpointName() != null
                && consumes.getServiceName() != null) {
            endpoint = component.getContext().getEndpoint(
                    consumes.getServiceName(), consumes.getEndpointName());
        }
        return endpoint;
    }

    /**
     * Return the operation according to the WSAAction SOAP attribute. If not
     * found in the WSA, get it from the SOAP body (root element)
     *
     * @param inContext
     *            SOAPrequest context
     * @return operation as QName, null if no WSAAction is specified
     */
    protected QName retrieveSOAPOperation(MessageContext inContext) {
        QName operationName = null;
        String operation = inContext.getWSAAction();
        if (!StringHelper.isNullOrEmpty(operation)) {
            operationName = QName.valueOf(operation);
        } else {
            operationName = retrieveOperationFromSOAPBody(inContext);
        }
        return operationName;
    }

    /**
     * Process the JBI response. <br>
     * According to the response status :
     * <li>
     * <ul>
     * do nothin for DONE,
     * </ul>
     * <ul>
     * throws an AxisFault for ERROR,
     * </ul>
     * <ul>
     * set outContext body to a FAULT,
     * </ul>
     * <ul>
     * set outContext body to the response.
     * </ul>
     * </li>
     *
     * @param exchange
     *            JBI response
     * @param outMessage
     *            set to FAULT or OUT JBI response
     * @param factory
     *            soap factory used to create body response
     * @throws AxisFault
     *             with ERROR JBI status, or problem while creating response
     */
    protected void processJBIResponse(MessageExchange exchange,
            MessageContext outMessage, SOAPFactory factory) throws AxisFault {
        if (exchange.getStatus().equals(ExchangeStatus.DONE)) {
            // exchange DONE, nothing to do, no soap response

        } else if (exchange.getStatus().equals(ExchangeStatus.ERROR)) {
            // exchange ERROR: throws an axis fault
            throw new AxisFault(FAULT_SERVER, exchange.getError());

        } else if (exchange.getStatus().equals(ExchangeStatus.ACTIVE)) {
            if (exchange.getFault() != null) {
                // exchange FAULT : return a fault message
                try {
                    SOAPEnvelope envelope = marshaller
                            .createSOAPEnvelopeFromNM(factory, exchange
                                    .getFault());
                    outMessage.setEnvelope(envelope);
                } catch (AxisFault e) {
                    String errorMsg = "Error while creating SOAP Fault response.";
                    logger.log(Level.SEVERE, errorMsg, e);
                    throw new AxisFault(errorMsg, FAULT_SERVER, e);
                }

            } else {
                // Get the output message if the message is an instance of out
                try {
                    NormalizedMessage nm = null;
                    if (exchange instanceof InOptionalOut) {
                        InOptionalOut mex = (InOptionalOut) exchange;
                        nm = mex.getOutMessage();
                    } else if (exchange instanceof InOut) {
                        InOut mex = (InOut) exchange;
                        nm = mex.getOutMessage();
                    }

                    SOAPEnvelope envelope = marshaller
                            .createSOAPEnvelopeFromNM(factory, nm);
                    outMessage.setEnvelope(envelope);

                    marshaller.copyJBIAttachmentsToSOAP(nm, outMessage);
                } catch (AxisFault e) {
                    String errorMsg = "Error while creating SOAP Fault response.";
                    logger.log(Level.SEVERE, errorMsg, e);
                    throw new AxisFault(errorMsg, FAULT_SERVER, e);
                }
            }
            // close the messageExchange with the JBI NMR
            try {
                exchange.setStatus(ExchangeStatus.DONE);
                channel.send(exchange);
            } catch (MessagingException e) {
                String errorMsg = "Error while closing JBI MessageExchange.";
                logger.log(Level.SEVERE, errorMsg, e);
                throw new AxisFault(errorMsg, FAULT_SERVER, e);
            }

        } else {
            throw new AxisFault(ERROR_WRONG_MESSAGE_STATUS + " "
                    + exchange.getStatus().toString());
        }
    }

    /**
     * Get the operation from the message context. Assume that the operation is
     * always the root element of the SOAP body.
     *
     * @param message
     * @return null if the operation has not been found
     */
    protected QName retrieveOperationFromSOAPBody(MessageContext message) {
        QName result = null;
        SOAPEnvelope env = message.getEnvelope();
        SOAPBody body = env.getBody();
        OMElement operation = body.getFirstElement();
        if (operation != null) {
            result = operation.getQName();
        }
        return result;
    }

    /**
     * An util method creating a message exchange matching a pattern specified
     * in CDKExtensions. Valid patterns are : "in-only", "robust-in-only",
     * "in-opt-out" or "in-out"
     *
     * @param extensions
     *            a {@link CDKExtensions}, must not be null
     * @return a message exchange matching the pattern given in extensions
     * @throws MessagingException
     *             if the message exchange can't be created, if the specified
     *             pattern isn't valid or if no pattern specified in extensions
     */
    protected MessageExchange createMessageExchange(
            final Extensions extensions) throws MessagingException {
        if (extensions == null) {
            throw new NullPointerException(
                    "'extensions' parameter must be non null");
        }
        MEPConstants mep = retrievePatternFromExtensions(extensions);
        return createMessageExchange(mep);
    }

    /**
     * Retrieve the MEP from the extensions.
     *
     * @param extensions
     * @return MEP specified in extensions or IN_OPTIONAL_OUT one if not found
     */
    protected final MEPConstants retrievePatternFromExtensions(
            final Extensions extensions) {
        MEPConstants mep = null;
        String patternShortcut = extensions.getPetalsExtensions().getProperty(
                ExtensionsConstants.MEP);

        if (!StringHelper.isNullOrEmpty(patternShortcut)) {
            if (AbstractComponent.ROBUST_IN_ONLY
                    .equalsIgnoreCase(patternShortcut)) {
                mep = MEPConstants.ROBUST_IN_ONLY_PATTERN;
            } else if (AbstractComponent.IN_OPTIONAL_OUT
                    .equalsIgnoreCase(patternShortcut)) {
                mep = MEPConstants.IN_OPTIONAL_OUT_PATTERN;
            } else if (AbstractComponent.IN_OUT
                    .equalsIgnoreCase(patternShortcut)) {
                mep = MEPConstants.IN_OUT_PATTERN;
            } else if (AbstractComponent.IN_ONLY
                    .equalsIgnoreCase(patternShortcut)) {
                mep = MEPConstants.IN_ONLY_PATTERN;
            } else {
                mep = MEPConstants.IN_OPTIONAL_OUT_PATTERN;
            }
        } else {
            mep = MEPConstants.IN_OPTIONAL_OUT_PATTERN;
        }
        return mep;
    }

    /**
     * Retrieve the MEP from the axisService
     *
     * @param operation
     * @param axisService
     * @return
     */
    protected final MEPConstants retrievePatternFromDefinition(QName operation,
            PetalsAxisService axisService) {
        MEPConstants mep = null;
        logger.fine("Get MEP for operation " + operation);

        try {
            // get from cache
            URI mepURI = axisService.getOperationMep().get(operation);
            if (mepURI == null) {
                // retrieve MEP from definition (done on first call)
                Document definition = axisService.getServiceDefinition();
                if (definition != null) {
                    // WsdlHelper.getOperation(definition, bodyName,
                    // serviceName,
                    // interfaceName, endpointName);
                }
            }
        } catch (Exception e) {
        }
        return mep;
    }

    /**
     * Get the pattern of the given operation. First, try to get pattern from
     * the WSDL definition of the service. If the result is null, try to get the
     * pattern from the extensions. Note that in this case, the pattern is not
     * defined for the operation but to all of them.
     *
     * @param operation
     * @param axisService
     * @return
     */
    protected final MEPConstants getPattern(QName operation,
            PetalsAxisService axisService) {
        MEPConstants mep = retrievePatternFromDefinition(operation, axisService);
        if (mep == null) {
            mep = retrievePatternFromExtensions(axisService.getConsumesCDKExtensions());
        }
        return mep;
    }

    /**
     * An utility method creating a message exchange matching the given pattern.
     *
     * @param mep
     *            a specific MEP pattern
     * @return a message exchange matching the given pattern
     * @throws MessagingException
     *             if the message exchange can't be created or if the specified
     *             pattern isn't valid
     */
    protected MessageExchange createMessageExchange(
            final MEPConstants mep) throws MessagingException {
        if (mep == null) {
            throw new IllegalArgumentException("'mep' parameter must be non null");
        }
        MessageExchange exchange = null;
        if (mep.equals(MEPConstants.IN_ONLY_PATTERN)) {
            exchange = channel.createExchangeFactory().createInOnlyExchange();
        } else if (mep.equals(MEPConstants.IN_OPTIONAL_OUT_PATTERN)) {
            exchange = channel.createExchangeFactory()
                    .createInOptionalOutExchange();
        } else if (mep.equals(MEPConstants.IN_OUT_PATTERN)) {
            exchange = channel.createExchangeFactory().createInOutExchange();
        } else if (mep.equals(MEPConstants.ROBUST_IN_ONLY_PATTERN)) {
            exchange = channel.createExchangeFactory()
                    .createRobustInOnlyExchange();
        } else {
            throw new MessagingException("Given mep doesn't exist");
        }
        return exchange;
    }
}
